/**
* Zeigt eine Programmieraufgabe zum Lösen an
*/
function TaskUI(
    nodeContainer       // DOM element to occupy
) {
    nodeContainer.innerHTML = '';
    this._msgPlaceGeneral = new ScoTools.MessagePlace(true, true);
    this._msgPlaceGeneral.appendAsChild(nodeContainer);
    this._msgPlace = new ScoTools.MessagePlace(false, true);
    this._msgPlace.appendAsChild(nodeContainer);

    this._contentContainer = ScoTools.createHtml('div', nodeContainer, { 'class': 'codingtask' });
    const divAuftrag = ScoTools.createHtml('div', this._contentContainer, { 'class': 'aufgabe' });
    this._taskHeading = ScoTools.createHtml('h2', divAuftrag, null, 'Auftrag');
    this._taskView = ScoTools.createHtml('div', divAuftrag);
    this._taskView.innerHTML = '<h2>Wird geladen…</h2>Diese Meldung sollte schnell verschwinden.';
    this._editor = CodeMirror.fromTextArea(ScoTools.createHtml('textarea', ScoTools.createHtml('div', this._contentContainer), {}, this._stdSetting.code), {
        lineNumbers: true,
        mode: "text/x-java",
        theme: "default",
        viewportMargin: Infinity,
        indentUnit: 4
    });
    this._editor.on("change", () => this._procChanged());

    this._btnSubmit = ScoTools.createHtml('button', nodeContainer, { 'class': 'fullwidth' },  '📥 Klassendefintion überprüfen');
    this._btnSubmit.addEventListener('click', () => this.submit());

    this._importTask();
}

// Verwendete ScoTools.MessagePlace-Instanz für generelle Meldungen
TaskUI.prototype._msgPlaceGeneral = null;

// Verwendete ScoTools.MessagePlace-Instanz für aktuelle Meldung
TaskUI.prototype._msgPlace = null;

// Verwendetes Element für Überschrift
TaskUI.prototype._taskHeading = null;

// Verwendeter Container für Aufgabenstellung
TaskUI.prototype._taskView = null;

// Verwendete CodeMirror-Instanz
TaskUI.prototype._editor = null;

// Verwendete Abgabe-Schaltfläche
TaskUI.prototype._btnSubmit = null;

// Container (DOM-Knoten) für Hauptinhalt
TaskUI.prototype._contentContainer = null;

// Aktueller Java-ähnlicher Quellcode
Object.defineProperty(TaskUI.prototype, 'code', {
    enumerable: true,
    set(code) { this._editor.setValue(code); },
    get() { return this._editor.getValue(); }
});

// Funktion, die nach jeder Veränderung ausgeführt werden soll
TaskUI.prototype.onchange = function() { }

/**
* Übergeordneter Instanz mitteilen, dass sich Daten verändert haben
* @return undefined
*/
TaskUI.prototype._procChanged = function() {
    if ((typeof this.onchange == 'function') || (this.onchange instanceof Function))
        this.onchange.call(this);
}

// Wurden Aufgabe bereits abgeschlossen?
TaskUI.prototype._done = false;
Object.defineProperty(TaskUI.prototype, 'done', {
    enumerable: true,
    get() { return this._done; },
    set(v) {
        if ((v ? true : false) <= this._done)
            return;
        this._done = true;
        this.addGeneralMsg(this.createFinalReport(), ScoTools.Message.TYPE_SUCCESS);
        this._procChanged();
    }
});

// Gesammelte Anzahl an Fehlerpunkten
TaskUI.prototype._fineCount = 0;
Object.defineProperty(TaskUI.prototype, 'fineCount', {
    enumerable: true,
    get() { return this._fineCount; },
    set(v) {
        if ((v == this._fineCount) || !Number.isFinite(v))
            return;

        const oldFine = this._fineCount;
        this._fineCount = v;
        this._procChanged();
    }
});

// Geladene Aufgabendaten
TaskUI.prototype._stdSetting = { code: '// Hier Klassendefintion einfügen\n' };

/**
* Daten von window.CODINGTASK übernehmen
* @return undefined
*/
TaskUI.prototype._importTask = function() {
    if (!(window.CODINGTASK instanceof Object))
        return;
    let data = window.CODINGTASK;
    if (data instanceof Function)
        data = window.TASK.call(window.TASK);
    if (!(data instanceof Object))
        return;
    if (data.code)
        this.code = data.code;
    if (data.labelSubmit)
        this._btnSubmit.textContent = data.labelSubmit;
    if (data.heading)
        this._taskHeading.textContent = data.heading;
    if (data.html)
        this._taskView.innerHTML = data.html;
    if (data.canReset) {
        const resetBtn = ScoTools.createHtml('button', ScoTools.createHtml('p', this._taskView.parentNode, { 'class': 'resetBtnPlace' }));
        ScoTools.createHtml('img', resetBtn, { alt: "⎌", src: "share/Edit-undo.svg" });
        ScoTools.createHtml(false, resetBtn, null, ' Auf Ausgangslage zurücksetzen');
        resetBtn.addEventListener('click', () => {
            if (confirm("Achtung: Aktueller Quelltext wird gelöscht und durch ursprünglichen Quelltext ersetzt."))
                this.reset();
        });
    }
    this._stdSetting = data;
}

/**
* Früher gespeicherte Werte wieder übernehmen
* @return undefined
*/
TaskUI.prototype.loadValues = function(
    isDone,                 // Wurde diese Aufgabe bereits abgeschlossen?
    fineCount,              // Anzahl der gesammelten Fehlerpunkte
    code                    // Eingegebener Quelltext
) {
    this._fineCount = (Number.isFinite(fineCount) ? fineCount : 0);
    if (isDone)
        this.done = true;
    if (code)
        this.code = code;
}

/**
* Die Speichertabelle und alles andere zurücksetzen
* @return undefined
*/
TaskUI.prototype.reset = function() {
    if (this._stdSetting.code !== undefined)
        this.code = this._stdSetting.code;
}

/**
* Abschlussbericht zusammenstellen
* @return String        HTML-Code des Abschlussberichts
*/
TaskUI.prototype.createFinalReport = function() {
    if (!this.fineCount)
        return 'Aufgabe fehlerfrei gelöst';
    if (this.fineCount == 1)
        return 'Aufgabe gelöst: Ein Fehlerpunkt';
    const arrZahlen = ['', '', 'Zwei', 'Drei', 'Vier', 'Fünf', 'Sechs', 'Sieben', 'Acht', 'Neun', 'Zehn', 'Elf', 'Zwölf'];
    if (this.fineCount < arrZahlen.length)
        return 'Aufgabe gelöst: ' + arrZahlen[this.fineCount] + ' Fehlerpunkte';
    return 'Aufgabe gelöst: ' + this.fineCount + ' Fehlerpunkte';
}

/**
* Auf fehlerhaften Code hinweisen
* @return undefined
*/
TaskUI.prototype.reportCodeError = function(
    desc,       // Beschreibung des Fehlers
    where       // Zugehörige JavaToken- oder JavaPart-Instanz
) {
    this.addMsg(desc);
    const from = { line: undefined, ch: undefined };
    const to = { line: undefined, ch: undefined };
    if (where instanceof JavaToken) {
        from.line = where.line - 1;
        from.ch = where.pos - 1;
        to.line = from.line;
        to.ch = from.ch + ((where.raw || where.value) + "").length;
    } else if ((where instanceof JavaPart) && (where.tokenStart instanceof JavaToken)) {
        from.line = where.tokenStart.line - 1;
        from.ch = where.tokenStart.pos - 1;
        to.line = (where.tokenEnd instanceof JavaToken) ? (where.tokenEnd.line - 1) : Infinity;
        to.ch = (where.tokenEnd instanceof JavaToken) ? (where.tokenEnd.pos - 1 + ((where.tokenEnd.raw || where.tokenEnd.value) + "").length) : Infinity;
    }
    if (to.line === undefined || (to.line == from.line) && (to.ch == from.line))
        this._editor.setCursor(from.line, from.ch);
    else
        this._editor.setSelection(from, to);
    this._editor.focus();
}

/**
* Eine (weitere) Meldung dauerhaft anzeigen
* @return undefined
*/
TaskUI.prototype.addGeneralMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._msgPlaceGeneral.add(text, type);
}

/**
* Aktive Meldung löschen
* @return undefined
*/
TaskUI.prototype.clearMsg = function() {
    this._msgPlace.clear();
}

/**
* Eine Meldung vorübergehend anzeigen
* @return undefined
*/
TaskUI.prototype.addMsg = function(
    text,       // Optional: Meldungstext (Standardwert undefined für bisherige generische Meldung löschen)
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._msgPlace.add(text, type);
}

/**
* Drücken der Überprüfen-Schaltfläche bearbeiten
* @return undefined
*/
TaskUI.prototype.submit = function() {
    this.check();
}

/**
* Fehlerpunkte vergeben
* @return undefined
*/
TaskUI.prototype.addFine = function(
    isSyntaxError       // Handelt es sich um einen Syntaxfehler?
) {
    const newFine = isSyntaxError
        ? (this._stdSetting.fineSyntax ? this._stdSetting.fineSyntax : 0)
        : (Number.isFinite(this._stdSetting.fineContent) ? this._stdSetting.fineContent : 1);
    if (newFine)
        this.fineCount += newFine;
}

/**
* Bearbeitung des aktuellen Schritts korrigieren
* @return Boolean       Schritt richtig bearbeitet?
*/
TaskUI.prototype.check = function() {
    this.clearMsg();

    if (!(this._stdSetting.checkData instanceof Array) && this.done) {
        this.addMsg("Abgabe kann nicht mehr verändert werden", ScoTools.MSGTYPE_WARN);
        return false;
    }

    const tokens = (new JavaLexer(this.code)).tokens;
    const [errorToken] = tokens.filter(token => !token.type);
    if (errorToken) {
        this.reportCodeError("Syntaxfehler: „" + errorToken.value + "“ nicht erkannt", errorToken);
        this.addFine(true);
        return false;
    }

    const relevantTokens = tokens.filter(token => ([JavaToken.TYPE_COMMENT, JavaToken.TYPE_WHITESPACE, JavaToken.TYPE_ERROR].indexOf(token.type) == -1));
    const classDefs = [];
    while (relevantTokens.length) {
        const warnings = [];
        const classDefData = JavaPartClassDeclaration.fromTokens(relevantTokens, warnings) || JavaPartImport.fromTokens(relevantTokens, warnings);
        if (warnings.length) {
            const { desc, token } = warnings[0];
            this.reportCodeError(desc, token);
            this.addFine(true);
            return false;
        }
        if (!classDefData) {
            this.reportCodeError("Mit „" + relevantTokens[0].value + "“ kann ich hier nichts anfangen", relevantTokens[0]);
            this.addFine(true);
            return false;
        }
        classDefs.push(classDefData);
    }

    if (!(this._stdSetting.checkData instanceof Array)) {
        if (confirm("Der eingegebene Quelltext wird nun endgültig abgegeben und kann nicht mehr verändert werden.")) {
            this.done = true;
            return true;
        }
        this.addMsg("Abgabe abgebrochen", ScoTools.MSGTYPE_INFO);
        return false;
    }

    const result = JavaPart.compareContent(classDefs, this._stdSetting.checkData);
    if (result.problems.length) {
        this.reportCodeError(
            result.problems[0] + ((result.problems.length > 1) ? ', …' : ''),
            result.instance
        );
        this.addFine(false);
        return false;
    }
    this.done = true;
    return true;
}
